#ifndef _XC3S_I2C_H
#define _XC3S_I2C_H

#include "Prefix.h"
#include "OS_Semaphore.h"
#include "OS_Thread.h"
#include "XC3S_Register.h"

namespace EARTH {
namespace XC3S {
	class I2C {
	public:
		I2C(XC3S::Register *register_) : mRegister(register_)
		{
			_ASSERTE(mRegister);

			mSemaphore = NULL;
			mState = STATE_CLEAR;

			_ASSERTE(mSemaphore == NULL);
			mSemaphore = new OS::Semaphore(1);
			_ASSERTE(mSemaphore);
		}

		~I2C()
		{
			delete_(mSemaphore);
		}

		void Init()
		{
			_ASSERTE(mSemaphore);
			mSemaphore->Acquire();

			_ASSERTE(mState == STATE_CLEAR);

			code code;

			code.busy         = false;
			code.read_enable  = false;
			//code.read_index   = 0;
			code.clock        = 1;
			code.data         = 1;
			code.next_address = 0;

			for (uint i=0; i<1<<BIT_PROGRAM_ADDRESS; i++) {
				WriteCode(i, &code);
			}

			// Wait (512 PCI Clocks)
			for (uint i=0; i<(512)/4; i++) {
				WriteCode(0, &code);
			}

			mState = STATE_STOP;

			mSemaphore->Release();
		}

		bool Write(const byte *ptr, uint size = 1)
		{
			_ASSERTE(mSemaphore);
			mSemaphore->Acquire();

			_ASSERTE(mState != STATE_CLEAR);

			bool b = WriteBlock(ptr, size);
			_ASSERTE(b);

			mSemaphore->Release();

			return b;
		}

		bool Read(const byte *writePtr, uint writeSize, byte *readPtr, uint readSize = 1)
		{
			_ASSERTE(mSemaphore);
			mSemaphore->Acquire();

			_ASSERTE(mState != STATE_CLEAR);

			bool b = ReadBlock(writePtr, writeSize, readPtr, readSize);

			mSemaphore->Release();

			return b;
		}

	protected:
		enum {
			A_WRITE_CONTROL = 0x00,
			A_WRITE_CODE    = 0x04,
			A_READ_BUSY     = 0x00,
			A_READ_DATA     = 0x02
		};

		enum {
			BIT_WRITE_I2C_BEGIN = 2
		};

		enum {
			BIT_PROGRAM_ADDRESS = 10,
			BIT_PROGRAM_DATA    = 18
		};

		enum {
			BIT_WRITE_PROGRAM_DATA,
			BIT_WRITE_PROGRAM_ADDRESS = BIT_WRITE_PROGRAM_DATA + BIT_PROGRAM_DATA
		};

		enum {
			BIT_READ_BUSY = 7
		};

		enum {
			MAX_WRITE_SIZE = (1024 - 3 - 4) / (3 * 9) - 2
		};

		enum state {
			STATE_CLEAR,
			STATE_STOP,
			STATE_NORMAL
		};

		XC3S::Register	*const mRegister;
		OS::Semaphore	*mSemaphore;
		state			mState;

		struct code {
			bool	busy;
			bool	read_enable;
			//uint	read_index;
			uint	clock,
					data;
			uint	next_address;
		};

		bool WriteBlock(const byte *ptr, uint size)
		{
			_ASSERTE(ptr);
			_ASSERTE((1 <= size) && (size <= MAX_WRITE_SIZE));

			code	code;
			uint	address = 0;

			Begin(&address, &code);

			if (mState == STATE_STOP) {
				Start(&address, &code);
				Stop (&address, &code, false);
				mState = STATE_NORMAL;
			}

			Start(&address, &code);

			for (uint i=0; i<size; i++) {
				uint	data = ptr[i] << 1 | BIT_1(1);

				for (uint j=0; j<9; j++) {
					code.clock        = 0;
					code.next_address = address+1;
					WriteCode(address++, &code);

					uint data_bit = BIT_SHIFT_MASK(data, 8-j, 1);

					code.data         = data_bit;
					code.next_address = address+1;
					WriteCode(address++, &code);

					code.clock        = 1;
					code.next_address = address+1;
					WriteCode(address++, &code);
				}
			}

			// Clock negedge
			code.clock        = 0;
			code.next_address = address+1;
			WriteCode(address++, &code);

			Stop(&address, &code);

			mRegister->Write4(A_WRITE_CONTROL, BIT_1(1) << BIT_WRITE_I2C_BEGIN);

			uint data;
			while (true) {
				OS::Thread::Sleep((1)*1000*1000);

				mRegister->Read4(A_READ_BUSY, &data);

				if (!BIT_SHIFT_MASK(data, BIT_READ_BUSY, 1)) break;
			}

			return true;	
		}

		bool ReadBlock(const byte *writePtr, uint writeSize, byte *readPtr, uint readSize)
		{
			code	code;
			uint	address = 0;

			Begin(&address, &code);

			if (mState == STATE_STOP) {
				Start(&address, &code);
				Stop (&address, &code, false);
				mState = STATE_NORMAL;
			}

			Start(&address, &code);

			for (uint i=0; i<writeSize+readSize; i++) {
				uint	data;
				if (i < writeSize) {
					data = writePtr[i] << 1 | BIT_1(1);
				} else {
					data = BIT_8(1,1,1,1,1,1,1,1) << 1 | BIT_1(i == writeSize+readSize-1);
				}

				for (uint j=0; j<9; j++) {
					code.clock        = 0;
					code.next_address = address+1;
					WriteCode(address++, &code);

					uint data_bit = BIT_SHIFT_MASK(data, 8-j, 1);

					code.data         = data_bit;
					code.next_address = address+1;
					WriteCode(address++, &code);

					code.clock        = 1;
					code.next_address = address+1;
					WriteCode(address++, &code);

					if ((writeSize <= i) && (j < 8)) {
						code.read_enable  = true;
						code.next_address = address+1;
						WriteCode(address++, &code);

						code.read_enable  = false;
					}

					if ((i == writeSize-2) && (j == 8)) {
						// Clock negedge
						code.clock        = 0;
						code.next_address = address+1;
						WriteCode(address++, &code);

						Start(&address, &code);
					}
				}
			}

			// Clock negedge
			code.clock        = 0;
			code.next_address = address+1;
			WriteCode(address++, &code);

			Stop(&address, &code);

			mRegister->Write4(A_WRITE_CONTROL, BIT_1(1) << BIT_WRITE_I2C_BEGIN);

			uint data;
			while (true) {
				OS::Thread::Sleep((1)*1000*1000);

				mRegister->Read4(A_READ_BUSY, &data);

				if (!BIT_SHIFT_MASK(data, BIT_READ_BUSY, 1)) break;
			}

			mRegister->Read4(A_READ_DATA, &data);
			for (uint i=0; i<readSize; i++) {
				readPtr[i] = static_cast<byte>(data >> 8*(readSize-1-i));
			}

			return true;	
		}

		void Begin(uint *address_, code *code)
		{
			_ASSERTE(address_);
			uint	address = *address_;

			_ASSERTE(code);
			code->busy         = false;
			code->read_enable  = false;
			//code->read_index   = 0;
			code->clock        = 1;
			code->data         = 1;
			code->next_address = address+0;
			WriteCode(address++, code);

			code->busy         = true;
			code->next_address = address+1;
			WriteCode(address++, code);

			*address_ = address;
		}

		void Start(uint *address_, code *code)
		{
			_ASSERTE(address_);
			uint	address = *address_;

			_ASSERTE(code);
			if (code->data != 1) {
				if (code->clock != 0) {
					code->clock        = 0;
					code->next_address = address+1;
					WriteCode(address++, code);
				}

				code->data         = 1;
				code->next_address = address+1;
				WriteCode(address++, code);
			}

			if (code->clock != 1) {
				code->clock        = 1;
				code->next_address = address+1;
				WriteCode(address++, code);
			}

			code->data         = 0;
			code->next_address = address+1;
			WriteCode(address++, code);

			*address_ = address;
		}

		void Stop(uint *address_, code *code, bool end = true)
		{
			_ASSERTE(address_);
			uint	address = *address_;

			_ASSERTE(code);
			if (code->data != 0) {
				if (code->clock != 0) {
					code->clock        = 0;
					code->next_address = address+1;
					WriteCode(address++, code);
				}

				code->data         = 0;
				code->next_address = address+1;
				WriteCode(address++, code);
			}

			if (code->clock != 1) {
				code->clock        = 1;
				code->next_address = address+1;
				WriteCode(address++, code);
			}

			code->data         = 1;
			code->next_address = end ? 0 : address+1;
			WriteCode(address++, code);

			*address_ = address;
		}

		uint Code(const code *code_)
		{
			_ASSERTE(code_);
			code	code = *code_;

			code.clock        &= 1;
			code.data         &= 1;
			code.next_address &= (1 << BIT_PROGRAM_ADDRESS) - 1;

			enum {
				BIT_NEXT_ADDRESS,
				BIT_DATA = BIT_NEXT_ADDRESS + BIT_PROGRAM_ADDRESS,
				BIT_CLOCK,
				BIT_READ_ENABLE,
				BIT_BUSY
			};

			uint	data =
				   code.next_address         << BIT_NEXT_ADDRESS
				| (code.data        ? 0 : 1) << BIT_DATA
				| (code.clock       ? 0 : 1) << BIT_CLOCK
				| (code.read_enable ? 1 : 0) << BIT_READ_ENABLE
				| (code.busy        ? 1 : 0) << BIT_BUSY;

			return data;
		}

		void WriteCode(uint address, const code *code_)
		{
			_ASSERTE(code_);
			uint	code = Code(code_);

			uint	data =
				  (address << BIT_WRITE_PROGRAM_ADDRESS)
				| (code    << BIT_WRITE_PROGRAM_DATA   );


			mRegister->Write4(A_WRITE_CODE, data);
		}
};

}
}

#endif
