#ifndef _XC3S_PowerReset_H
#define _XC3S_PowerReset_H

#include "Prefix.h"
#include "EARTH_PT.h"
#include "XC3S_Register.h"
#include "PCI_Device.h"
#include "OS_Thread.h"

namespace EARTH {
namespace XC3S {
	class Register;

	class PowerReset {
	public:
		PowerReset(PCI::Device *device, XC3S::Register *register_) : mDevice(device), mRegister(register_)
		{
			_ASSERTE(mDevice);
			_ASSERTE(mRegister);

			mLnbPower = mLnbPowerWhenClose = PT::Device::LNB_POWER_OFF;
			mTunerPowerReset = PT::Device::TUNER_POWER_OFF;

			for (uint i=0; i<2; i++) {
				for (uint j=0; j<PT::Device::ISDB_COUNT; j++) {
					mTunerSleep[i][j] = true;
				}
			}
		}

		~PowerReset()
		{
		}

		status SetLnbPower(PT::Device::LnbPower lnbPower)
		{
			mLnbPower = lnbPower;
			mRegister->Write4(A_WRITE, Data());

			return PT::STATUS_OK;
		}

		status GetLnbPower(PT::Device::LnbPower *lnbPower) const
		{
			if (lnbPower == NULL) return PT::STATUS_INVALID_PARAM_ERROR;

			*lnbPower = mLnbPower;

			return PT::STATUS_OK;
		}

		status SetLnbPowerWhenClose(PT::Device::LnbPower lnbPower)
		{
			status status;

			uint data = LnbPowerReset(lnbPower, PT::Device::TUNER_POWER_OFF);
			status = mDevice->SetReset(A_WRITE, data);
			if (status) return status;

			mLnbPowerWhenClose = lnbPower;

			return status;
		}

		status GetLnbPowerWhenClose(PT::Device::LnbPower *lnbPower) const
		{
			if (lnbPower == NULL) return PT::STATUS_INVALID_PARAM_ERROR;

			*lnbPower = mLnbPowerWhenClose;

			return PT::STATUS_OK;
		}

		status SetTunerPowerReset(PT::Device::TunerPowerReset tunerPowerReset)
		{
			mTunerPowerReset = tunerPowerReset;

			mRegister->Write4(A_WRITE, Data());

			return PT::STATUS_OK;
		}

		status GetTunerPowerReset(PT::Device::TunerPowerReset *tunerPowerReset) const
		{
			if (tunerPowerReset == NULL) return PT::STATUS_INVALID_PARAM_ERROR;

			*tunerPowerReset = mTunerPowerReset;

			return PT::STATUS_OK;
		}

		status SetTunerSleep(uint tuner, PT::Device::ISDB isdb, bool sleep)
		{
			mTunerSleep[tuner][isdb] = sleep;

			mRegister->Write4(A_WRITE, Data());

			return PT::STATUS_OK;
		}

	protected:
		enum {
			A_WRITE = 0x01
		};

		PCI::Device *const mDevice;
		XC3S::Register *const mRegister;
		PT::Device::LnbPower mLnbPower, mLnbPowerWhenClose;
		PT::Device::TunerPowerReset mTunerPowerReset;
		bool mTunerSleep[2][PT::Device::ISDB_COUNT];

		uint LnbPowerReset(PT::Device::LnbPower lnbPower, PT::Device::TunerPowerReset tunerPowerReset) const
		{
			enum {
				BIT_TUNER =  0,
				BIT_LNB_UP,
				BIT_LNB_DOWN,
				BIT_RESET
			};

			uint data = 0;

			switch (lnbPower) {
				case PT::Device::LNB_POWER_OFF: data |= BIT_1(0) << BIT_LNB_UP | BIT_1(0) << BIT_LNB_DOWN; break;
				case PT::Device::LNB_POWER_15V: data |= BIT_1(1) << BIT_LNB_UP | BIT_1(1) << BIT_LNB_DOWN; break;
				case PT::Device::LNB_POWER_11V: data |= BIT_1(0) << BIT_LNB_UP | BIT_1(1) << BIT_LNB_DOWN; break;
				default                       : return PT::STATUS_GENERAL_ERROR;
			}

			switch (tunerPowerReset) {
				case PT::Device::TUNER_POWER_OFF             : data |= BIT_1(0) << BIT_TUNER | BIT_1(0) << BIT_RESET; break;
				case PT::Device::TUNER_POWER_ON_RESET_ENABLE : data |= BIT_1(1) << BIT_TUNER | BIT_1(0) << BIT_RESET; break;
				case PT::Device::TUNER_POWER_ON_RESET_DISABLE: data |= BIT_1(1) << BIT_TUNER | BIT_1(1) << BIT_RESET; break;
				default                                      : return PT::STATUS_GENERAL_ERROR;
			}

			return data;
		}

		uint LnbPowerReset() const
		{
			return LnbPowerReset(mLnbPower, mTunerPowerReset);
		}

		uint Sleep() const
		{
			enum {
				BIT_33A1 = 4,
				BIT_33A2,
				BIT_5A_,
				BIT_5A1,
				BIT_5A2
			};

			const uint DATA[2][PT::Device::ISDB_COUNT] = {
				BIT_1(1) << BIT_33A1,
				BIT_1(1) << BIT_5A1 | BIT_1(1) << BIT_5A_,
				BIT_1(1) << BIT_33A2,
				BIT_1(1) << BIT_5A2 | BIT_1(1) << BIT_5A_
			};

			uint data = 0;

			for (uint i=0; i<2; i++) {
				for (uint j=0; j<PT::Device::ISDB_COUNT; j++) {
					if (mTunerSleep[i][j] == false) {
						data |= DATA[i][j];
					}
				}
			}

			return data;
		}

		uint Data() const
		{
			return LnbPowerReset() | Sleep();
		}
	};
}
}

#endif
