#ifndef _PCI_Device__H
#define _PCI_Device__H

#include "PCI_Device.h"
#include "PCI.h"
#include "EARTH_PT.h"
#include <windows.h>
#include "windrvr.h"

#if defined(KERNEL_32BIT) || defined(KERNEL_64BIT)

namespace EARTH {
namespace PCI {
namespace KERNEL_BIT {

	using namespace EARTH;

	class Device_ : public Device {
	public:	
		Device_(HANDLE handle, const Target *target) : mHandle(handle)
		{
			_ASSERTE(mHandle != INVALID_HANDLE_VALUE);

			mBusItem = -1;
			mMemoryItem = -1;
			BZERO(mReg);
			mNeedsUnregister = false;
			
			for (uint i=0; i<sizeof(mData)/sizeof(*mData); i++) {
				mData[i] = 0;
			}

			_ASSERTE(target);
			mSlot.dwBus      = target->Bus;
			mSlot.dwSlot     = target->Slot;
			mSlot.dwFunction = target->Function;
		}

		virtual ~Device_()
		{
			if (mNeedsUnregister) {
				_ASSERTE(mHandle != INVALID_HANDLE_VALUE);
				DWORD status = WD_CardUnregister(mHandle, &mReg);
				_ASSERTE(status == WD_STATUS_SUCCESS);
				mNeedsUnregister = false;
			}
		}

		status Init()
		{
			WD_PCI_CARD_INFO card;
			BZERO(card);
			card.pciSlot = mSlot;

			status status;

			status = WD_PciGetCardInfo(mHandle, &card);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_PciGetCardInfo_ERROR;

			for (uint i=0; i<card.Card.dwItems; i++) {
				const WD_ITEMS *item = &card.Card.Item[i];
				_ASSERTE(item);

				switch (item->item) {
					case ITEM_BUS:
					{
						if (0 <= mBusItem) return PT::STATUS_WD_PciGetCardInfo_Bus_ERROR;
						mBusItem = i;
						break;
					}
					case ITEM_MEMORY:
					{
						if (0 <= mMemoryItem) return PT::STATUS_WD_PciGetCardInfo_Memory_ERROR;
						mMemoryItem = i;
						break;
					}
				}
			}

			if (mBusItem < 0) return PT::STATUS_WD_PciGetCardInfo_Bus_ERROR;
			if (mMemoryItem < 0) return PT::STATUS_WD_PciGetCardInfo_Memory_ERROR;

			BZERO(mReg);
			mReg.Card            = card.Card;
			mReg.fCheckLockOnly  = FALSE;
			mReg.dwOptions       = 0;
			mReg.cName[0]        = 0;
			mReg.cDescription[0] = 0;


			status = WD_CardRegister(mHandle, &mReg);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_CardRegister_ERROR;
			mNeedsUnregister = true;

			return PT::STATUS_OK;
		}

		virtual void Delete()
		{
			delete this;
		}

		virtual status WriteConfig(uint address, const void *data, uint size)
		{
			WD_PCI_CONFIG_DUMP config;
			BZERO(config);
			config.pciSlot  = mSlot;
			config.pBuffer  = const_cast<void *>(data);
			config.dwOffset = address;
			config.dwBytes  = size;
			config.fIsRead  = FALSE;

			DWORD status = WD_PciConfigDump(mHandle, &config);
			if ((status != WD_STATUS_SUCCESS) || (config.dwResult != PCI_ACCESS_OK)) return PT::STATUS_WD_PciConfigDump_ERROR;

			return PT::STATUS_OK;
		}

		virtual status ReadConfig(uint address, void *data, uint size)
		{
			WD_PCI_CONFIG_DUMP config;
			BZERO(config);
			config.pciSlot  = mSlot;
			config.pBuffer  = data;
			config.dwOffset = address;
			config.dwBytes  = size;
			config.fIsRead  = TRUE;

			DWORD status = WD_PciConfigDump(mHandle, &config);
			if ((status != WD_STATUS_SUCCESS) || (config.dwResult != PCI_ACCESS_OK)) return PT::STATUS_WD_PciConfigDump_ERROR;

			return PT::STATUS_OK;
		}

		virtual void WriteMemory(uint address, const void *data, uint size)
		{
			volatile byte *ptr = reinterpret_cast<byte *>(mReg.Card.Item[mMemoryItem].I.Mem.dwUserDirectAddr);
			_ASSERTE(ptr);

			switch (size) {
				case 1 : *reinterpret_cast<volatile byte   *>(ptr + address) = *static_cast<const byte   *>(data); break;
				case 2 : *reinterpret_cast<volatile ushort *>(ptr + address) = *static_cast<const ushort *>(data); break;
				case 4 : *reinterpret_cast<volatile uint   *>(ptr + address) = *static_cast<const uint   *>(data); break;
				default: _ASSERTE(false);                                                                          break;
			}
		}

		virtual void ReadMemory(uint address, void *data, uint size)
		{
			volatile byte *ptr = reinterpret_cast<byte *>(mReg.Card.Item[mMemoryItem].I.Mem.dwUserDirectAddr);
			_ASSERTE(ptr);

			switch (size) {
				case 1 : *static_cast<byte   *>(data) = *reinterpret_cast<volatile byte   *>(ptr + address); break;
				case 2 : *static_cast<ushort *>(data) = *reinterpret_cast<volatile ushort *>(ptr + address); break;
				case 4 : *static_cast<uint   *>(data) = *reinterpret_cast<volatile uint   *>(ptr + address); break;
				default: _ASSERTE(false);                                                                    break;
			}
		}

		virtual status SetReset(uint address, uint data, bool callWinDriver)
		{
			_ASSERTE(address < sizeof(mData)/sizeof(*mData));
			mData[address] = data;
			if (callWinDriver == false) return PT::STATUS_OK;

			WD_TRANSFER transfer[2];

			for (uint i=0; i<sizeof(transfer)/sizeof(*transfer); i++) {
				WD_TRANSFER *ptr = &transfer[i];
				ptr->dwPort     = mReg.Card.Item[mMemoryItem].I.Mem.dwTransAddr + 4*i;
				ptr->cmdTrans   = WM_DWORD;
				ptr->dwBytes    = 4;
				ptr->fAutoinc   = FALSE;
				ptr->dwOptions  = 0;
				ptr->Data.Dword = mData[i];
			}

			WD_CARD_CLEANUP clean;
			BZERO(clean);

			clean.hCard     = mReg.hCard;
			clean.Cmd       = transfer;
			clean.dwCmds    = sizeof(transfer)/sizeof(*transfer);
			clean.dwOptions = WD_FORCE_CLEANUP;

			uint status = WD_CardCleanupSetup(mHandle, &clean);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_CardCleanupSetup_ERROR;

			return PT::STATUS_OK;
		}

		virtual uint ID() const
		{
			return mReg.hCard;
		}

	protected:
		const HANDLE mHandle;
		WD_PCI_SLOT mSlot;
		int mBusItem, mMemoryItem;
		WD_CARD_REGISTER mReg;
		bool mNeedsUnregister;
		uint mData[2];
	};

	status NewDevice(HANDLE handle, const Target *target, Device **device_)
	{
		_ASSERTE(target);
		Device *device = new Device_(handle, target);
		status status = device->Init();
		if (status) {
			device->Delete();
			device = NULL;
		}

		_ASSERTE(device_);
		*device_ = device;

		return status;
	}
}
}
}

#endif

#endif
